from __future__ import annotations

import asyncio
import contextlib
import logging
import math
from typing import TYPE_CHECKING, Any

import anyio
from anyio import create_memory_object_stream

from waku.messaging.context import MessageContext, reset_message_context, set_message_context
from waku.messaging.contracts.envelope import MessageEnvelope
from waku.messaging.dispatcher import MessageDispatcher
from waku.messaging.endpoints.base import Endpoint, HandlerSubscriptions

if TYPE_CHECKING:
    from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
    from dishka import AsyncContainer

logger = logging.getLogger(__name__)


class LocalQueueEndpoint(Endpoint):
    __slots__ = ('_container', '_receive_stream', '_send_stream', '_stop_timeout', '_worker_task')

    def __init__(
        self,
        *,
        uri: str,
        handler_subscriptions: HandlerSubscriptions,
        container: AsyncContainer,
        stop_timeout: float,
    ) -> None:
        super().__init__(uri=uri, handler_subscriptions=handler_subscriptions)
        self._container = container
        self._stop_timeout = stop_timeout
        send, receive = create_memory_object_stream[MessageEnvelope[Any]](max_buffer_size=math.inf)
        self._send_stream: MemoryObjectSendStream[MessageEnvelope[Any]] = send
        self._receive_stream: MemoryObjectReceiveStream[MessageEnvelope[Any]] = receive
        self._worker_task: asyncio.Task[None] | None = None

    async def dispatch(self, envelope: MessageEnvelope[Any], scope: AsyncContainer) -> None:  # noqa: ARG002
        await self._send_stream.send(envelope)

    async def start(self) -> None:
        self._worker_task = asyncio.create_task(self._worker_loop())

    async def stop(self) -> None:
        self._send_stream.close()
        if self._worker_task is None:
            return
        try:
            with anyio.fail_after(self._stop_timeout):
                await self._worker_task
        except TimeoutError:
            logger.warning(
                'Worker task for %s did not terminate within %.1fs, cancelling',
                self._uri,
                self._stop_timeout,
            )
            self._worker_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._worker_task
        finally:
            self._worker_task = None

    async def _worker_loop(self) -> None:
        async with self._receive_stream:
            async for envelope in self._receive_stream:
                await self._process_envelope(envelope)

    async def _process_envelope(self, envelope: MessageEnvelope[Any]) -> None:
        try:
            async with self._container() as scope:
                dispatcher = await scope.get(MessageDispatcher)
                ctx = MessageContext(
                    correlation_id=envelope.correlation_id,
                    causation_id=envelope.causation_id,
                    message_id=envelope.message_id,
                    headers=envelope.headers,
                )
                token = set_message_context(ctx)
                try:
                    await self._execute(dispatcher, envelope)
                finally:
                    reset_message_context(token)
        except Exception:
            logger.exception(
                'Failed to process message: message_id=%s, message_type=%s',
                envelope.message_id,
                envelope.message_type,
            )

    async def _execute(self, dispatcher: MessageDispatcher, envelope: MessageEnvelope[Any]) -> None:
        payload = envelope.payload
        handler_types = self._handler_subscriptions.get(type(payload))
        if handler_types is not None:
            await dispatcher.publish_event_only(payload, only=handler_types)
        else:
            await dispatcher.invoke_request(payload)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, overload

from typing_extensions import override

from waku.di import AsyncContainer  # noqa: TC001  # Dishka needs runtime access
from waku.messaging.context import (
    MessageContext,
    reset_message_context,
    set_message_context,
    try_get_message_context,
)
from waku.messaging.contracts.factory import EnvelopeFactory  # noqa: TC001  # Dishka needs runtime access
from waku.messaging.dispatcher import MessageDispatcher  # noqa: TC001  # Dishka needs runtime access
from waku.messaging.interfaces import IMessageBus
from waku.messaging.router import MessageRouter  # noqa: TC001  # Dishka needs runtime access

if TYPE_CHECKING:
    from contextvars import Token

    from waku.messaging.contracts.envelope import MessageEnvelope
    from waku.messaging.contracts.event import IEvent
    from waku.messaging.contracts.message import ResponseT
    from waku.messaging.contracts.request import IRequest


class MessageBus(IMessageBus):
    __slots__ = ('_container', '_dispatcher', '_envelope_factory', '_router')

    def __init__(
        self,
        container: AsyncContainer,
        dispatcher: MessageDispatcher,
        envelope_factory: EnvelopeFactory,
        router: MessageRouter,
    ) -> None:
        self._container = container
        self._dispatcher = dispatcher
        self._envelope_factory = envelope_factory
        self._router = router

    @overload
    async def invoke(self, request: IRequest[None], /) -> None: ...

    @overload
    async def invoke(self, request: IRequest[ResponseT], /) -> ResponseT: ...

    @override
    async def invoke(self, request: IRequest[Any], /) -> Any:
        envelope = self._create_envelope(request)
        token = self._set_context(envelope)
        try:
            return await self._dispatcher.invoke_request(request)
        finally:
            self._reset_context(token)

    @override
    async def send(self, request: IRequest[Any], /) -> None:
        envelope = self._create_envelope(request)
        endpoints = self._router.resolve(type(request))
        if not endpoints:
            token = self._set_context(envelope)
            try:
                await self._dispatcher.invoke_request(request)
            finally:
                self._reset_context(token)
        else:
            for endpoint in endpoints:
                await endpoint.dispatch(envelope, self._container)

    @override
    async def publish(self, event: IEvent, /) -> None:
        envelope = self._create_envelope(event)
        endpoints = self._router.resolve(type(event))
        if not endpoints:
            token = self._set_context(envelope)
            try:
                await self._dispatcher.publish_event(event)
            finally:
                self._reset_context(token)
        else:
            excluded = self._router.routed_handler_types(type(event))
            token = self._set_context(envelope)
            try:
                await self._dispatcher.publish_event_excluding(event, exclude=excluded)
            finally:
                self._reset_context(token)
            for endpoint in endpoints:
                await endpoint.dispatch(envelope, self._container)

    def _create_envelope(self, message: Any) -> MessageEnvelope[Any]:
        ctx = try_get_message_context()
        if ctx is not None:
            return self._envelope_factory.create(
                message,
                correlation_id=ctx.correlation_id,
                causation_id=ctx.message_id,
            )
        return self._envelope_factory.create(message)

    @staticmethod
    def _set_context(envelope: MessageEnvelope[Any]) -> Token[MessageContext | None]:
        ctx = MessageContext(
            correlation_id=envelope.correlation_id,
            causation_id=envelope.causation_id,
            message_id=envelope.message_id,
            headers=envelope.headers,
        )
        return set_message_context(ctx)

    @staticmethod
    def _reset_context(token: Token[MessageContext | None]) -> None:
        reset_message_context(token)



from collections import defaultdict
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Self, TypeAlias

from typing_extensions import override

from waku.di import AsyncContainer, Provider, WithParents, many, object_, scoped, singleton, transient
from waku.extensions import AfterApplicationInit, OnApplicationShutdown, OnModuleConfigure, OnModuleRegistration
from waku.messaging.context import MessageContext, get_message_context
from waku.messaging.contracts.event import EventT
from waku.messaging.contracts.factory import EnvelopeFactory
from waku.messaging.contracts.pipeline import IPipelineBehavior
from waku.messaging.contracts.request import RequestT
from waku.messaging.dispatcher import MessageDispatcher
from waku.messaging.endpoints.base import Endpoint, EndpointEntry, EndpointKind
from waku.messaging.endpoints.local_queue import LocalQueueEndpoint
from waku.messaging.events.handler import EventHandler
from waku.messaging.impl import MessageBus
from waku.messaging.interfaces import IMessageBus
from waku.messaging.pipeline.map import PipelineBehaviorMapEntry
from waku.messaging.registry import MessageRegistry
from waku.messaging.requests.handler import RequestHandler
from waku.messaging.router import MessageRouter, ModuleRouteDescriptor, RouteDescriptor, RoutingTable
from waku.messaging.routing_builder import RoutingTableBuilder
from waku.modules import DynamicModule, ModuleMetadataRegistry, module

if TYPE_CHECKING:
    from waku.application import WakuApplication
    from waku.messaging.contracts.event import IEvent
    from waku.messaging.contracts.message import IMessage
    from waku.modules import ModuleMetadata, ModuleType

__all__ = [
    'MessagingConfig',
    'MessagingExtension',
    'MessagingModule',
]


_HandlerProviders: TypeAlias = tuple[Provider, ...]


@dataclass(frozen=True, slots=True, kw_only=True)
class MessagingConfig:
    """Configuration for the messaging extension.

    Attributes:
        pipeline_behaviors: A sequence of pipeline behavior configurations that will be applied
            to the messaging pipeline. Behaviors are executed in the order they are defined.
            Defaults to an empty sequence.
        endpoints: A sequence of endpoint entries defining available message endpoints.
            Defaults to an empty sequence.
        routing: A sequence of route descriptors mapping messages to endpoints.
            Defaults to an empty sequence.

    Example:
        ```python
        config = MessagingConfig(
            pipeline_behaviors=[
                LoggingBehavior,
                ValidationBehavior,
            ]
        )
        ```
    """

    pipeline_behaviors: Sequence[type[IPipelineBehavior[Any, Any]]] = ()
    endpoints: Sequence[EndpointEntry] = ()
    routing: Sequence[RouteDescriptor | ModuleRouteDescriptor] = ()


@module()
class MessagingModule:
    @classmethod
    def register(cls, config: MessagingConfig | None = None, /) -> DynamicModule:
        """Application-level module for MessageBus setup.

        Args:
            config: Configuration for the messaging extension.
        """
        config_ = config or MessagingConfig()
        return DynamicModule(
            parent_module=cls,
            providers=[
                scoped(WithParents[IMessageBus], MessageBus),  # ty:ignore[not-subscriptable]
                singleton(EnvelopeFactory),
                scoped(MessageDispatcher),
                transient(MessageContext, get_message_context),
                *cls._create_pipeline_behavior_providers(config_),
            ],
            extensions=[
                MessageRegistryAggregator(config_),
                EndpointLifecycleExtension(),
            ],
            is_global=True,
        )

    @staticmethod
    def _create_pipeline_behavior_providers(config: MessagingConfig) -> _HandlerProviders:
        if not config.pipeline_behaviors:
            return ()
        return (many(IPipelineBehavior[Any, Any], *config.pipeline_behaviors),)


class MessagingExtension(OnModuleConfigure):
    def __init__(self) -> None:
        self._registry = MessageRegistry()

    @override
    def on_module_configure(self, metadata: 'ModuleMetadata') -> None:
        pass

    def bind_request(
        self,
        request_type: type[RequestT],
        handler_type: type[RequestHandler[RequestT, Any]],
        *,
        behaviors: list[type[IPipelineBehavior[RequestT, Any]]] | None = None,
    ) -> Self:
        self._registry.request_map.bind(request_type, handler_type)
        if behaviors:
            request_entry: PipelineBehaviorMapEntry[Any, Any] = PipelineBehaviorMapEntry.for_request(request_type)
            self._registry.behavior_map.bind(request_entry, behaviors)
        return self

    def bind_event(
        self,
        event_type: type[EventT],
        handler_types: list[type[EventHandler[EventT]]],
        *,
        behaviors: list[type[IPipelineBehavior[EventT, None]]] | None = None,
    ) -> Self:
        self._registry.event_map.bind(event_type, handler_types)
        if behaviors:
            event_entry: PipelineBehaviorMapEntry[Any, Any] = PipelineBehaviorMapEntry.for_event(event_type)
            self._registry.behavior_map.bind(event_entry, behaviors)
        return self

    @property
    def registry(self) -> MessageRegistry:
        return self._registry


def _create_router(routing_table: RoutingTable, container: AsyncContainer) -> MessageRouter:
    endpoints_by_uri: dict[str, Endpoint] = {}
    for entry in routing_table.entries:
        if entry.kind == EndpointKind.LOCAL_QUEUE:
            endpoints_by_uri[entry.uri] = LocalQueueEndpoint(
                uri=entry.uri,
                handler_subscriptions=dict(entry.handler_subscriptions),
                container=container,
                stop_timeout=entry.stop_timeout,
            )

    routes: defaultdict[type[IMessage], list[Endpoint]] = defaultdict(list)
    for msg_type, uris in routing_table.type_routes.items():
        for uri in uris:
            if uri in endpoints_by_uri:
                routes[msg_type].append(endpoints_by_uri[uri])

    return MessageRouter(
        routes=routes,
        handler_routes=dict(routing_table.handler_routes),
        endpoints=list(endpoints_by_uri.values()),
    )


class MessageRegistryAggregator(OnModuleRegistration):
    __slots__ = ('_config',)

    def __init__(self, config: MessagingConfig) -> None:
        self._config = config

    @override
    def on_module_registration(
        self,
        registry: ModuleMetadataRegistry,
        owning_module: 'ModuleType',
        context: Mapping[Any, Any] | None,
    ) -> None:
        aggregated = MessageRegistry()
        module_routing_map: dict[type, dict[type[IEvent], list[type[EventHandler[Any]]]]] = {}

        for module_type, ext in registry.find_extensions(MessagingExtension):
            aggregated.merge(ext.registry)
            event_handlers: dict[type[IEvent], list[type[EventHandler[Any]]]] = {}
            for entry in ext.registry.event_map.entries():
                event_handlers[entry.event_type] = list(entry.handler_types)
            if event_handlers:
                module_routing_map[module_type] = event_handlers
            for provider in ext.registry.handler_providers():
                registry.add_provider(module_type, provider)

        for provider in aggregated.collector_providers():
            registry.add_provider(owning_module, provider)

        aggregated.freeze()
        registry.add_provider(owning_module, object_(aggregated))

        routing_table = RoutingTableBuilder(
            self._config,
            aggregated=aggregated,
            module_routing_map=module_routing_map,
        ).build()
        registry.add_provider(owning_module, object_(routing_table))
        registry.add_provider(owning_module, singleton(MessageRouter, _create_router))


class EndpointLifecycleExtension(AfterApplicationInit, OnApplicationShutdown):
    __slots__ = ()

    @override
    async def after_app_init(self, app: 'WakuApplication') -> None:
        router = await app.container.get(MessageRouter)
        for endpoint in router.endpoints:
            await endpoint.start()

    @override
    async def on_app_shutdown(self, app: 'WakuApplication') -> None:
        router = await app.container.get(MessageRouter)
        for endpoint in reversed(router.endpoints):
            await endpoint.stop()



from __future__ import annotations

from typing_extensions import TypeVar

__all__ = [
    'IMessage',
    'MessageT',
    'ResponseT',
]


class IMessage:
    __slots__ = ()


MessageT = TypeVar('MessageT', bound=IMessage, contravariant=True)  # noqa: PLC0105
ResponseT = TypeVar('ResponseT', default=None, covariant=True)  # noqa: PLC0105



from __future__ import annotations

from datetime import UTC, datetime
from typing import TYPE_CHECKING, TypeVar
from uuid import uuid4

from waku.messaging.contracts.envelope import MessageEnvelope

if TYPE_CHECKING:
    from collections.abc import Mapping
    from uuid import UUID

T = TypeVar('T')


class EnvelopeFactory:
    @classmethod
    def create(
        cls,
        message: T,
        *,
        message_id: UUID | None = None,
        correlation_id: UUID | None = None,
        causation_id: UUID | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> MessageEnvelope[T]:
        message_id_ = message_id or uuid4()
        return MessageEnvelope(
            message_id=message_id_,
            correlation_id=correlation_id or uuid4(),
            causation_id=causation_id or message_id_,
            message_type=f'{type(message).__module__}.{type(message).__qualname__}',
            timestamp=datetime.now(tz=UTC),
            payload=message,
            headers=headers or {},
        )



from __future__ import annotations

from typing import Protocol, runtime_checkable


@runtime_checkable
class IUnitOfWork(Protocol):
    async def commit(self) -> None: ...
    async def rollback(self) -> None: ...



from __future__ import annotations

from dataclasses import dataclass, field
from types import MappingProxyType
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from waku.messaging.contracts.message import IMessage
    from waku.messaging.endpoints.base import Endpoint, EndpointEntry
    from waku.messaging.events.handler import EventHandler

_EMPTY_TYPE_ROUTES: Mapping[type[IMessage], tuple[str, ...]] = MappingProxyType({})
_EMPTY_HANDLER_ROUTES: Mapping[type[IMessage], frozenset[type[EventHandler[Any]]]] = MappingProxyType({})


@dataclass(frozen=True, slots=True)
class RoutingTable:
    entries: tuple[EndpointEntry, ...] = ()
    type_routes: Mapping[type[IMessage], tuple[str, ...]] = field(default_factory=lambda: _EMPTY_TYPE_ROUTES)
    handler_routes: Mapping[type[IMessage], frozenset[type[EventHandler[Any]]]] = field(
        default_factory=lambda: _EMPTY_HANDLER_ROUTES
    )


class MessageRouter:
    __slots__ = ('_endpoints', '_handler_routes', '_routes')

    def __init__(
        self,
        routes: dict[type[IMessage], list[Endpoint]],
        handler_routes: dict[type[IMessage], frozenset[type[EventHandler[Any]]]],
        endpoints: Sequence[Endpoint] = (),
    ) -> None:
        self._routes: dict[type[IMessage], tuple[Endpoint, ...]] = {k: tuple(v) for k, v in routes.items()}
        self._handler_routes = handler_routes
        self._endpoints = tuple(endpoints)

    @property
    def endpoints(self) -> tuple[Endpoint, ...]:
        return self._endpoints

    def resolve(self, message_type: type[IMessage]) -> tuple[Endpoint, ...]:
        return self._routes.get(message_type, ())

    def routed_handler_types(self, message_type: type[IMessage]) -> frozenset[type[EventHandler[Any]]]:
        return self._handler_routes.get(message_type, frozenset())


@dataclass(frozen=True, slots=True)
class RouteDescriptor:
    message_type: type[IMessage]
    endpoint_uri: str


@dataclass(frozen=True, slots=True)
class ModuleRouteDescriptor:
    module_type: type
    endpoint_uri: str


class RouteBuilder:
    __slots__ = ('_message_type',)

    def __init__(self, message_type: type[IMessage]) -> None:
        self._message_type = message_type

    def to(self, endpoint_uri: str) -> RouteDescriptor:
        return RouteDescriptor(self._message_type, endpoint_uri)


class ModuleRouteBuilder:
    __slots__ = ('_module_type',)

    def __init__(self, module_type: type) -> None:
        self._module_type = module_type

    def events_to(self, endpoint_uri: str) -> ModuleRouteDescriptor:
        return ModuleRouteDescriptor(self._module_type, endpoint_uri)


def route(message_type: type[IMessage]) -> RouteBuilder:
    return RouteBuilder(message_type)


def route_module(module_type: type) -> ModuleRouteBuilder:
    return ModuleRouteBuilder(module_type)



from __future__ import annotations

import abc
from typing import Generic

from waku.messaging.contracts.message import ResponseT
from waku.messaging.contracts.request import RequestT

__all__ = [
    'RequestHandler',
]


class RequestHandler(abc.ABC, Generic[RequestT, ResponseT]):
    """Abstract base class for request handlers (commands/queries).

    Command handler example::

        class JoinMeetingCommandHandler(RequestHandler[JoinMeetingCommand, None]):
            def __init__(self, meetings_api: MeetingAPIProtocol) -> None:
                self._meetings_api = meetings_api

            async def handle(self, request: JoinMeetingCommand, /) -> None:
                await self._meetings_api.join_user(request.user_id, request.meeting_id)

    Query handler example::

        class ReadMeetingQueryHandler(RequestHandler[ReadMeetingQuery, ReadMeetingQueryResult]):
            def __init__(self, meetings_api: MeetingAPIProtocol) -> None:
                self._meetings_api = meetings_api

            async def handle(self, request: ReadMeetingQuery, /) -> ReadMeetingQueryResult:
                link = await self._meetings_api.get_link(request.meeting_id)
                return ReadMeetingQueryResult(link=link, meeting_id=request.meeting_id)

    """

    @abc.abstractmethod
    async def handle(self, request: RequestT, /) -> ResponseT:
        raise NotImplementedError



from waku.messaging.contracts.event import EventT, IEvent
from waku.messaging.contracts.message import IMessage, MessageT, ResponseT
from waku.messaging.contracts.pipeline import CallNext, IPipelineBehavior
from waku.messaging.contracts.request import IRequest, RequestT

__all__ = [
    'CallNext',
    'EventT',
    'IEvent',
    'IMessage',
    'IPipelineBehavior',
    'IRequest',
    'MessageT',
    'RequestT',
    'ResponseT',
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from collections.abc import Mapping
    from datetime import datetime
    from uuid import UUID

T = TypeVar('T')


@dataclass(frozen=True, slots=True, kw_only=True)
class MessageEnvelope(Generic[T]):
    message_id: UUID
    correlation_id: UUID
    causation_id: UUID
    message_type: str
    timestamp: datetime
    payload: T
    headers: Mapping[str, str] = field(default_factory=dict)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

if TYPE_CHECKING:
    from waku.messaging.contracts.envelope import MessageEnvelope


@runtime_checkable
class ITransport(Protocol):
    async def send(self, envelope: MessageEnvelope[Any]) -> None: ...
    async def publish(self, envelope: MessageEnvelope[Any]) -> None: ...



from __future__ import annotations

import abc
from typing import Generic

from waku.messaging.contracts.event import EventT

__all__ = [
    'EventHandler',
]


class EventHandler(abc.ABC, Generic[EventT]):
    """Abstract base class for event handlers.

    Example::

        class UserJoinedEventHandler(EventHandler[UserJoinedEvent]):
            def __init__(self, meetings_api: MeetingAPIProtocol) -> None:
                self._meetings_api = meetings_api

            async def handle(self, event: UserJoinedEvent, /) -> None:
                await self._meetings_api.notify_room(event.meeting_id, 'New user joined!')

    """

    @abc.abstractmethod
    async def handle(self, event: EventT, /) -> None:
        raise NotImplementedError



from __future__ import annotations

from collections.abc import Iterator, MutableMapping, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Generic, Self, TypeAlias

from waku.messaging._introspection import get_request_response_type
from waku.messaging.contracts.message import IMessage, MessageT, ResponseT
from waku.messaging.contracts.pipeline import IPipelineBehavior
from waku.messaging.exceptions import MapFrozenError, PipelineBehaviorAlreadyRegistered

if TYPE_CHECKING:
    from waku.messaging.contracts.event import IEvent
    from waku.messaging.contracts.request import IRequest

__all__ = [
    'PipelineBehaviorMap',
    'PipelineBehaviorMapEntry',
]


@dataclass(slots=True)
class PipelineBehaviorMapEntry(Generic[MessageT, ResponseT]):
    message_type: type[IMessage]
    di_lookup_type: type[IPipelineBehavior[MessageT, ResponseT]]
    behavior_types: list[type[IPipelineBehavior[Any, Any]]] = field(default_factory=list)

    @classmethod
    def for_request(cls, request_type: type[IRequest[ResponseT]]) -> Self:
        response_type = get_request_response_type(request_type)
        di_lookup_type = IPipelineBehavior[request_type, response_type]  # type: ignore[valid-type]
        return cls(message_type=request_type, di_lookup_type=di_lookup_type)  # type: ignore[type-abstract]

    @classmethod
    def for_event(cls, event_type: type[IEvent]) -> Self:
        di_lookup_type = IPipelineBehavior[event_type, None]  # type: ignore[valid-type]
        return cls(message_type=event_type, di_lookup_type=di_lookup_type)  # type: ignore[type-abstract]

    def add(self, behavior_type: type[IPipelineBehavior[Any, Any]]) -> None:
        if behavior_type in self.behavior_types:
            raise PipelineBehaviorAlreadyRegistered(self.message_type, behavior_type)
        self.behavior_types.append(behavior_type)


PipelineBehaviorMapRegistry: TypeAlias = MutableMapping[type[MessageT], PipelineBehaviorMapEntry[MessageT, ResponseT]]


class PipelineBehaviorMap:
    def __init__(self) -> None:
        self._registry: PipelineBehaviorMapRegistry[Any, Any] = {}
        self._frozen = False

    def freeze(self) -> None:
        self._frozen = True

    @property
    def is_frozen(self) -> bool:
        return self._frozen

    def bind(
        self,
        entry: PipelineBehaviorMapEntry[Any, Any],
        behavior_types: list[type[IPipelineBehavior[Any, Any]]],
    ) -> Self:
        if self._frozen:
            raise MapFrozenError
        if entry.message_type not in self._registry:
            self._registry[entry.message_type] = entry

        existing = self._registry[entry.message_type]
        for behavior_type in behavior_types:
            existing.add(behavior_type)
        return self

    def merge(self, other: PipelineBehaviorMap) -> Self:
        if self._frozen:
            raise MapFrozenError
        for other_entry in other.entries():
            if other_entry.message_type not in self._registry:
                self._registry[other_entry.message_type] = PipelineBehaviorMapEntry(
                    message_type=other_entry.message_type,
                    di_lookup_type=other_entry.di_lookup_type,
                )
            target = self._registry[other_entry.message_type]
            for behavior_type in other_entry.behavior_types:
                target.add(behavior_type)
        return self

    def entries(self) -> Iterator[PipelineBehaviorMapEntry[Any, Any]]:
        yield from self._registry.values()

    def has_behaviors(self, message_type: type[IMessage]) -> bool:
        return message_type in self._registry and len(self._registry[message_type].behavior_types) > 0

    def get_behavior_types(self, message_type: type[IMessage]) -> Sequence[type[IPipelineBehavior[Any, Any]]]:
        return self._registry[message_type].behavior_types

    def get_lookup_type(self, message_type: type[IMessage]) -> type[IPipelineBehavior[Any, Any]]:
        return self._registry[message_type].di_lookup_type

    def __bool__(self) -> bool:
        return bool(self._registry)



from __future__ import annotations

from typing import TYPE_CHECKING

from waku.uow import IUnitOfWork

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession


class SqlAlchemyUnitOfWork(IUnitOfWork):
    __slots__ = ('_session',)

    def __init__(self, session: AsyncSession) -> None:
        self._session = session

    async def commit(self) -> None:
        await self._session.commit()

    async def rollback(self) -> None:
        await self._session.rollback()



from waku.messaging.contracts.event import EventT, IEvent
from waku.messaging.contracts.message import IMessage, MessageT, ResponseT
from waku.messaging.contracts.pipeline import CallNext, IPipelineBehavior
from waku.messaging.contracts.request import IRequest, RequestT
from waku.messaging.events.handler import EventHandler
from waku.messaging.impl import MessageBus
from waku.messaging.interfaces import IMessageBus, IPublisher, ISender
from waku.messaging.modules import MessagingConfig, MessagingExtension, MessagingModule
from waku.messaging.requests.handler import RequestHandler

__all__ = [
    'CallNext',
    'EventHandler',
    'EventT',
    'IEvent',
    'IMessage',
    'IMessageBus',
    'IPipelineBehavior',
    'IPublisher',
    'IRequest',
    'ISender',
    'MessageBus',
    'MessageT',
    'MessagingConfig',
    'MessagingExtension',
    'MessagingModule',
    'RequestHandler',
    'RequestT',
    'ResponseT',
]



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from typing import Generic, TypeAlias

from waku.messaging.contracts.message import MessageT, ResponseT

__all__ = [
    'CallNext',
    'IPipelineBehavior',
]

CallNext: TypeAlias = Callable[[], Awaitable[ResponseT]]


class IPipelineBehavior(ABC, Generic[MessageT, ResponseT]):
    @abstractmethod
    async def handle(
        self,
        message: MessageT,
        /,
        call_next: CallNext[ResponseT],
    ) -> ResponseT: ...



from __future__ import annotations

import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, TypeAlias

if TYPE_CHECKING:
    from collections.abc import Mapping

    from dishka import AsyncContainer

    from waku.messaging.contracts.envelope import MessageEnvelope
    from waku.messaging.contracts.message import IMessage
    from waku.messaging.events.handler import EventHandler

HandlerSubscriptions: TypeAlias = 'Mapping[type[IMessage], frozenset[type[EventHandler[Any]]]]'

_EMPTY_SUBSCRIPTIONS: HandlerSubscriptions = MappingProxyType({})


class EndpointKind(enum.StrEnum):
    LOCAL_QUEUE = 'local_queue'
    EXTERNAL = 'external'


@dataclass(frozen=True, slots=True)
class EndpointEntry:
    uri: str
    kind: EndpointKind
    handler_subscriptions: HandlerSubscriptions = field(default_factory=lambda: _EMPTY_SUBSCRIPTIONS)
    stop_timeout: float = 5.0


def local_queue(uri: str, *, stop_timeout: float = 5.0) -> EndpointEntry:
    return EndpointEntry(uri=uri, kind=EndpointKind.LOCAL_QUEUE, stop_timeout=stop_timeout)


def external_endpoint(uri: str) -> EndpointEntry:
    return EndpointEntry(uri=uri, kind=EndpointKind.EXTERNAL)


class Endpoint(ABC):
    __slots__ = ('_handler_subscriptions', '_uri')

    def __init__(self, uri: str, handler_subscriptions: HandlerSubscriptions) -> None:
        self._uri = uri
        self._handler_subscriptions = handler_subscriptions

    @property
    def uri(self) -> str:
        return self._uri

    @property
    def handler_subscriptions(self) -> HandlerSubscriptions:
        return self._handler_subscriptions

    @abstractmethod
    async def dispatch(self, envelope: MessageEnvelope[Any], scope: AsyncContainer) -> None: ...

    @abstractmethod
    async def start(self) -> None: ...

    @abstractmethod
    async def stop(self) -> None: ...



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast

from dishka.exceptions import NoFactoryError

from waku.di import AsyncContainer  # noqa: TC001  # Dishka needs runtime access
from waku.messaging.contracts.pipeline import IPipelineBehavior
from waku.messaging.exceptions import RequestHandlerNotFound
from waku.messaging.pipeline import PipelineExecutor
from waku.messaging.registry import MessageRegistry  # noqa: TC001  # Dishka needs runtime access

if TYPE_CHECKING:
    from waku.messaging.contracts.event import IEvent
    from waku.messaging.contracts.message import IMessage, ResponseT
    from waku.messaging.contracts.request import IRequest
    from waku.messaging.events.handler import EventHandler
    from waku.messaging.requests.handler import RequestHandler


class MessageDispatcher:
    __slots__ = ('_container', '_registry')

    def __init__(
        self,
        container: AsyncContainer,
        registry: MessageRegistry,
    ) -> None:
        self._container = container
        self._registry = registry

    async def invoke_request(self, request: IRequest[ResponseT]) -> ResponseT:
        request_type = type(request)
        handler = await self._resolve_request_handler(request_type)
        behaviors = await self._resolve_behaviors(request_type)
        return await PipelineExecutor.execute(message=request, handler=handler, behaviors=behaviors)  # pyrefly: ignore[bad-return]

    async def publish_event(self, event: IEvent) -> None:
        event_type = type(event)
        handlers = await self._resolve_event_handlers(event_type)
        behaviors = await self._resolve_behaviors(event_type)
        for h in handlers:
            await PipelineExecutor.execute(message=event, handler=h, behaviors=behaviors)

    async def publish_event_excluding(self, event: IEvent, *, exclude: frozenset[type[EventHandler[Any]]]) -> None:
        event_type = type(event)
        handlers = await self._resolve_event_handlers(event_type)
        behaviors = await self._resolve_behaviors(event_type)
        for h in handlers:
            if type(h) not in exclude:
                await PipelineExecutor.execute(message=event, handler=h, behaviors=behaviors)

    async def publish_event_only(self, event: IEvent, *, only: frozenset[type[EventHandler[Any]]]) -> None:
        event_type = type(event)
        handlers = await self._resolve_event_handlers(event_type)
        behaviors = await self._resolve_behaviors(event_type)
        for h in handlers:
            if type(h) in only:
                await PipelineExecutor.execute(message=event, handler=h, behaviors=behaviors)

    async def _resolve_request_handler(
        self,
        request_type: type[IRequest[ResponseT]],
    ) -> RequestHandler[IRequest[ResponseT], ResponseT]:
        if not self._registry.request_map.has_handler(request_type):
            raise RequestHandlerNotFound(request_type)

        handler_type = self._registry.request_map.get_handler_type(request_type)
        return cast('RequestHandler[IRequest[ResponseT], ResponseT]', await self._container.get(handler_type))

    async def _resolve_behaviors(self, message_type: type[IMessage]) -> Sequence[IPipelineBehavior[Any, Any]]:
        try:
            global_behaviors = await self._container.get(Sequence[IPipelineBehavior[Any, Any]])
        except NoFactoryError:
            global_behaviors = ()

        if not self._registry.behavior_map.has_behaviors(message_type):
            return global_behaviors

        lookup_type = self._registry.behavior_map.get_lookup_type(message_type)
        scoped_behaviors = await self._container.get(Sequence[lookup_type])  # type: ignore[valid-type]

        return (*global_behaviors, *scoped_behaviors)

    async def _resolve_event_handlers(
        self,
        event_type: type[IEvent],
    ) -> Sequence[EventHandler[IEvent]]:
        if not self._registry.event_map.has_handlers(event_type):
            return ()

        handler_type = self._registry.event_map.get_handler_type(event_type)
        handlers = await self._container.get(Sequence[handler_type])  # type: ignore[valid-type]
        return cast('Sequence[EventHandler[IEvent]]', handlers)  # pyrefly: ignore[redundant-cast]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from waku.messaging.contracts.pipeline import CallNext, IPipelineBehavior

if TYPE_CHECKING:
    from waku.uow import IUnitOfWork

logger = logging.getLogger(__name__)


class TransactionalBehavior(IPipelineBehavior[Any, Any]):
    __slots__ = ('_uow',)

    def __init__(self, uow: IUnitOfWork) -> None:
        self._uow = uow

    async def handle(self, _message: Any, /, call_next: CallNext[Any]) -> Any:
        try:
            result = await call_next()
        except Exception:
            await self._safe_rollback()
            raise
        try:
            await self._uow.commit()
        except Exception:
            await self._safe_rollback()
            raise
        return result

    async def _safe_rollback(self) -> None:
        try:
            await self._uow.rollback()
        except Exception:
            logger.exception('Rollback failed')



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, TypeVar

if TYPE_CHECKING:
    from collections.abc import Sequence

    from waku.messaging.contracts.message import IMessage
    from waku.messaging.contracts.pipeline import IPipelineBehavior

__all__ = [
    'PipelineExecutor',
]

_T_co = TypeVar('_T_co', covariant=True)
_T = TypeVar('_T')


class _MessageHandler(Protocol[_T_co]):
    async def handle(self, message: Any, /) -> _T_co: ...


class PipelineExecutor:
    @staticmethod
    async def execute(
        *,
        message: IMessage,
        handler: _MessageHandler[_T],
        behaviors: Sequence[IPipelineBehavior[Any, _T]],
    ) -> _T:
        async def terminal() -> _T:
            return await handler.handle(message)

        if not behaviors:
            return await terminal()

        async def step(idx: int) -> _T:
            if idx >= len(behaviors):
                return await terminal()
            return await behaviors[idx].handle(
                message,
                call_next=lambda: step(idx + 1),
            )

        return await step(0)



from __future__ import annotations

from contextvars import ContextVar, Token
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping
    from uuid import UUID


@dataclass(frozen=True, slots=True, kw_only=True)
class MessageContext:
    correlation_id: UUID
    causation_id: UUID
    message_id: UUID
    headers: Mapping[str, str]


_message_context: ContextVar[MessageContext | None] = ContextVar('_message_context', default=None)


def get_message_context() -> MessageContext:
    ctx = _message_context.get()
    if ctx is None:
        msg = 'No active message context. This function must be called within a MessageBus operation.'
        raise RuntimeError(msg)
    return ctx


def try_get_message_context() -> MessageContext | None:
    return _message_context.get()


def set_message_context(ctx: MessageContext) -> Token[MessageContext | None]:
    return _message_context.set(ctx)


def reset_message_context(token: Token[MessageContext | None]) -> None:
    _message_context.reset(token)



from __future__ import annotations

from collections import defaultdict
from collections.abc import Mapping, Sequence
from dataclasses import replace
from typing import TYPE_CHECKING, Any, assert_never

from waku.messaging.contracts.event import IEvent
from waku.messaging.exceptions import ImproperlyConfiguredError
from waku.messaging.router import ModuleRouteDescriptor, RouteDescriptor, RoutingTable

if TYPE_CHECKING:
    from waku.messaging.contracts.message import IMessage
    from waku.messaging.endpoints.base import EndpointEntry
    from waku.messaging.events.handler import EventHandler
    from waku.messaging.modules import MessagingConfig
    from waku.messaging.registry import MessageRegistry

__all__ = ['ModuleRoutingMap', 'RoutingTableBuilder']

ModuleRoutingMap = Mapping[type, Mapping[type[IEvent], Sequence[type['EventHandler[Any]']]]]


class RoutingTableBuilder:
    __slots__ = (
        '_config',
        '_handler_routes',
        '_module_routing_map',
        '_per_type_overrides',
        '_registry',
        '_type_routes',
    )

    def __init__(
        self,
        config: MessagingConfig,
        *,
        aggregated: MessageRegistry,
        module_routing_map: ModuleRoutingMap,
    ) -> None:
        self._config = config
        self._registry = aggregated
        self._module_routing_map = module_routing_map
        self._type_routes: defaultdict[type[IMessage], list[str]] = defaultdict(list)
        self._handler_routes: defaultdict[type[IMessage], set[type[EventHandler[Any]]]] = defaultdict(set)
        self._per_type_overrides: set[type[IMessage]] = set()

    def build(self) -> RoutingTable:
        endpoint_entries_by_uri = {entry.uri: entry for entry in self._config.endpoints}

        route_descriptors: list[RouteDescriptor] = []
        module_descriptors: list[ModuleRouteDescriptor] = []

        for descriptor in self._config.routing:
            self._validate_endpoint_uri(descriptor.endpoint_uri, endpoint_entries_by_uri)
            match descriptor:
                case RouteDescriptor():
                    route_descriptors.append(descriptor)
                case ModuleRouteDescriptor():
                    module_descriptors.append(descriptor)
                case _:
                    assert_never(descriptor)

        for descriptor in route_descriptors:
            self._process_route(descriptor)
        for descriptor in module_descriptors:
            self._process_module_route(descriptor)

        return self._assemble(endpoint_entries_by_uri)

    def _process_route(self, descriptor: RouteDescriptor) -> None:
        msg_type = descriptor.message_type
        self._per_type_overrides.add(msg_type)
        self._type_routes[msg_type].append(descriptor.endpoint_uri)

        if issubclass(msg_type, IEvent):
            self._handler_routes[msg_type].update(self._registry.event_map.get_handler_types(msg_type))

    def _process_module_route(self, descriptor: ModuleRouteDescriptor) -> None:
        module_type = descriptor.module_type
        if module_type not in self._module_routing_map:
            msg = f"route_module() references module '{module_type.__qualname__}' which has no registered events"
            raise ImproperlyConfiguredError(msg)

        for event_type, handler_types in self._module_routing_map[module_type].items():
            if event_type in self._per_type_overrides:
                continue
            self._type_routes[event_type].append(descriptor.endpoint_uri)
            self._handler_routes[event_type].update(handler_types)

    def _assemble(self, endpoint_entries_by_uri: dict[str, EndpointEntry]) -> RoutingTable:
        frozen_type_routes = {msg_type: tuple(dict.fromkeys(uris)) for msg_type, uris in self._type_routes.items()}
        frozen_handler_routes = {
            msg_type: frozenset(handlers) for msg_type, handlers in self._handler_routes.items() if handlers
        }

        enriched_entries: list[EndpointEntry] = []
        for uri, entry in endpoint_entries_by_uri.items():
            subs: dict[type[IMessage], frozenset[type[EventHandler[Any]]]] = {}
            for msg_type, uris in self._type_routes.items():
                if uri in uris and msg_type in frozen_handler_routes:
                    subs[msg_type] = frozen_handler_routes[msg_type]

            enriched_entries.append(replace(entry, handler_subscriptions=subs))

        return RoutingTable(
            entries=tuple(enriched_entries),
            type_routes=frozen_type_routes,
            handler_routes=frozen_handler_routes,
        )

    @staticmethod
    def _validate_endpoint_uri(uri: str, known: dict[str, EndpointEntry]) -> None:
        if uri not in known:
            msg = f"Route references unknown endpoint URI '{uri}'. Known endpoints: {sorted(known)}"
            raise ImproperlyConfiguredError(msg)
